build: Use .ONESHELL.
[paraslash.git] / sched.c
diff --git a/sched.c b/sched.c
index 1280e07e1040d475281d554ee8911c7afe8de613..a2903940fdaea1b24d6a49cfc2f54766c136070f 100644 (file)
--- a/sched.c
+++ b/sched.c
@@ -1,13 +1,8 @@
-/*
- * Copyright (C) 2006-2014 Andre Noll <maan@systemlinux.org>
- *
- * Licensed under the GPL v2. For licencing details see COPYING.
- */
+/* Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */
 
 /** \file sched.c Paraslash's scheduling functions. */
 
 #include <regex.h>
-#include <assert.h>
 
 #include "para.h"
 #include "ipc.h"
 #include "time.h"
 #include "error.h"
 
+/**
+ * The possible states of a task.
+ *
+ * In addition to the states listed here, a task may also enter zombie state.
+ * This happens when its ->post_select function returns negative, the ->status
+ * field is then set to this return value. Such tasks are not scheduled any
+ * more (i.e. ->pre_select() and ->post_select() are no longer called), but
+ * they stay on the scheduler task list until \ref task_reap() or
+ * \ref sched_shutdown() is called.
+ */
+enum task_status {
+       /** Task has been reaped and may be removed from the task list. */
+       TS_DEAD,
+       /** Task is active. */
+       TS_RUNNING,
+};
+
 struct task {
        /** A copy of the task name supplied when the task was registered. */
        char *name;
        /** Copied during task_register(). */
        struct task_info info;
-       /** Whether this task is active (>=0) or in error state (<0). */
+       /* TS_RUNNING, TS_DEAD, or zombie (negative value). */
        int status;
        /** Position of the task in the task list of the scheduler. */
        struct list_head node;
        /** If less than zero, the task was notified by another task. */
        int notification;
-       /** True if task is in error state and exit status has been queried. */
-       bool dead;
 };
 
 static struct timeval now_struct;
-struct timeval *now = &now_struct;
-
-static inline bool timeout_is_zero(struct sched *s)
-{
-       struct timeval *tv = &s->select_timeout;
-       return tv->tv_sec == 0 && tv->tv_usec == 0;
-}
+const struct timeval *now = &now_struct;
 
 static void sched_preselect(struct sched *s)
 {
@@ -52,13 +56,16 @@ static void sched_preselect(struct sched *s)
                if (t->notification != 0)
                        sched_min_delay(s);
                if (t->info.pre_select)
-                       t->info.pre_select(s, t);
+                       t->info.pre_select(s, t->info.context);
        }
 }
 
 static void unlink_and_free_task(struct task *t)
 {
-       PARA_INFO_LOG("freeing task %s\n", t->name);
+       PARA_INFO_LOG("freeing task %s (%s)\n", t->name, t->status < 0?
+               para_strerror(-t->status) :
+               (t->status == TS_DEAD? "[dead]" : "[running]"));
+
        list_del(&t->node);
        free(t->name);
        free(t);
@@ -67,14 +74,16 @@ static void unlink_and_free_task(struct task *t)
 //#define SCHED_DEBUG 1
 static inline void call_post_select(struct sched *s, struct task *t)
 {
+       int ret;
+
 #ifndef SCHED_DEBUG
-       t->status = t->info.post_select(s, t);
+       ret = t->info.post_select(s, t->info.context);
 #else
        struct timeval t1, t2, diff;
        unsigned long pst;
 
        clock_get_realtime(&t1);
-       t->status = t->info.post_select(s, t);
+       ret = t->info.post_select(s, t->info.context);
        clock_get_realtime(&t2);
        tv_diff(&t1, &t2, &diff);
        pst = tv2ms(&diff);
@@ -82,6 +91,7 @@ static inline void call_post_select(struct sched *s, struct task *t)
                PARA_WARNING_LOG("%s: post_select time: %lums\n",
                        t->name, pst);
 #endif
+       t->status = ret < 0? ret : TS_RUNNING;
 }
 
 static unsigned sched_post_select(struct sched *s)
@@ -90,15 +100,14 @@ static unsigned sched_post_select(struct sched *s)
        unsigned num_running_tasks = 0;
 
        list_for_each_entry_safe(t, tmp, &s->task_list, node) {
-               if (t->status < 0) {
-                       if (t->dead) /* task has been reaped */
-                               unlink_and_free_task(t);
-                       continue;
+               if (t->status == TS_DEAD) /* task has been reaped */
+                       unlink_and_free_task(t);
+               else if (t->status == TS_RUNNING) {
+                       call_post_select(s, t); /* sets t->status */
+                       t->notification = 0;
+                       if (t->status == TS_RUNNING)
+                               num_running_tasks++;
                }
-               call_post_select(s, t);
-               t->notification = 0;
-               if (t->status >= 0)
-                       num_running_tasks++;
        }
        return num_running_tasks;
 }
@@ -116,7 +125,7 @@ static unsigned sched_post_select(struct sched *s)
  * \return Zero if no more tasks are left in the task list, negative if the
  * select function returned an error.
  *
- * \sa \ref task, \ref now.
+ * \sa \ref now.
  */
 int schedule(struct sched *s)
 {
@@ -130,7 +139,7 @@ again:
        FD_ZERO(&s->wfds);
        s->select_timeout = s->default_timeout;
        s->max_fileno = -1;
-       clock_get_realtime(now);
+       clock_get_realtime(&now_struct);
        sched_preselect(s);
        ret = s->select_function(s->max_fileno + 1, &s->rfds, &s->wfds,
                &s->select_timeout);
@@ -146,7 +155,7 @@ again:
                FD_ZERO(&s->rfds);
                FD_ZERO(&s->wfds);
        }
-       clock_get_realtime(now);
+       clock_get_realtime(&now_struct);
        num_running_tasks = sched_post_select(s);
        if (num_running_tasks == 0)
                return 0;
@@ -159,7 +168,7 @@ again:
  * \param tptr Identifies the task to reap.
  *
  * This function is similar to wait(2) in that it returns information about a
- * terminated task and allows to release the resources associated with the
+ * terminated task which allows releasing the resources associated with the
  * task. Until this function is called, the terminated task remains in a zombie
  * state.
  *
@@ -175,6 +184,7 @@ again:
 int task_reap(struct task **tptr)
 {
        struct task *t;
+       int ret;
 
        if (!tptr)
                return 0;
@@ -183,19 +193,19 @@ int task_reap(struct task **tptr)
                return 0;
        if (t->status >= 0)
                return 0;
-       if (t->dead) /* will be freed in sched_post_select() */
-               return 0;
+       ret = t->status;
        /*
         * With list_for_each_entry_safe() it is only safe to remove the
         * _current_ list item. Since we are being called from the loop in
         * schedule() via some task's ->post_select() function, freeing the
         * given task here would result in use-after-free bugs in schedule().
* So we only set t->dead which tells schedule() to free the task in
* the next iteration of its loop.
         */
-       t->dead = true;
+       t->status = TS_DEAD;
+
        *tptr = NULL;
-       return t->status;
+       return ret;
 }
 
 /**
@@ -210,7 +220,7 @@ void sched_shutdown(struct sched *s)
        struct task *t, *tmp;
 
        list_for_each_entry_safe(t, tmp, &s->task_list, node) {
-               if (t->status >= 0)
+               if (t->status == TS_RUNNING)
                        /* The task list should contain only terminated tasks. */
                        PARA_WARNING_LOG("shutting down running task %s\n",
                                t->name);
@@ -239,25 +249,11 @@ struct task *task_register(struct task_info *info, struct sched *s)
        t->info = *info;
        t->name = para_strdup(info->name);
        t->notification = 0;
-       t->status = 0;
-       t->dead = false;
+       t->status = TS_RUNNING;
        list_add_tail(&t->node, &s->task_list);
        return t;
 }
 
-/**
- * Obtain the context pointer of a task.
- *
- * \param t Return this task's context pointer.
- *
- * \return A pointer to the memory location specified previously as \a
- * task_info->context when the task was registered with \ref task_register().
- */
-void *task_context(struct task *t)
-{
-       return t->info.context;
-}
-
 /**
  * Get the list of all registered tasks.
  *
@@ -276,7 +272,8 @@ char *get_task_list(struct sched *s)
        list_for_each_entry_safe(t, tmp, &s->task_list, node) {
                char *tmp_msg;
                tmp_msg = make_message("%s%p\t%s\t%s\n", msg? msg : "", t,
-                       t->status < 0? (t->dead? "dead" : "zombie") : "running",
+                       t->status == TS_DEAD? "dead" :
+                               (t->status == TS_RUNNING? "running" : "zombie"),
                        t->name);
                free(msg);
                msg = tmp_msg;
@@ -340,9 +337,9 @@ int task_status(const struct task *t)
 {
        if (!t)
                return 0;
-       if (t->dead)
+       if (t->status == TS_DEAD) /* pretend dead tasks don't exist */
                return 0;
-       if (t->status >= 0)
+       if (t->status == TS_RUNNING)
                return 1;
        return t->status;
 }
@@ -386,7 +383,7 @@ void sched_min_delay(struct sched *s)
  * function does nothing. Otherwise the timeout for the next select() call is
  * set to the given value.
  *
- * \sa sched_request_timeout_ms().
+ * \sa \ref sched_request_timeout_ms().
  */
 void sched_request_timeout(struct timeval *to, struct sched *s)
 {
@@ -419,7 +416,7 @@ void sched_request_timeout_ms(long unsigned ms, struct sched *s)
  * \return If \a barrier is in the past, this function does nothing and returns
  * zero. Otherwise it returns one.
  *
- * \sa sched_request_barrier_or_min_delay().
+ * \sa \ref sched_request_barrier_or_min_delay().
  */
 int sched_request_barrier(struct timeval *barrier, struct sched *s)
 {
@@ -440,7 +437,7 @@ int sched_request_barrier(struct timeval *barrier, struct sched *s)
  * \return If \a barrier is in the past, this function requests a minimal
  * timeout and returns zero. Otherwise it returns one.
  *
- * \sa sched_min_delay(), sched_request_barrier().
+ * \sa \ref sched_min_delay(), \ref sched_request_barrier().
  */
 int sched_request_barrier_or_min_delay(struct timeval *barrier, struct sched *s)
 {
@@ -183,19 +193,19 @@ int task_reap(struct task **tptr)
         * schedule() via some task's ->post_select() function, freeing the
         * given task here would result in use-after-free bugs in schedule().
-        * So we only set t->dead which tells schedule() to free the task in
-        * the next iteration of its loop.
+        * So we only set the task status to TS_DEAD which tells schedule() to
+        * free the task in the next iteration of its loop.
         */